package ru.dbotthepony.mc.otm.config

import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.math.defineDecimal
import ru.dbotthepony.mc.otm.registry.MNames

object ItemsConfig : AbstractConfig("items") {
	private fun verboseValues(name: String, storage: Decimal, receive: Decimal, extract: Decimal = receive): VerboseBalanceValues {
		builder.push(name)

		val obj = object : VerboseBalanceValues {
			override val capacity: Decimal by builder.defineDecimal("capacity", storage, minimum = Decimal.ONE)
			override val receive: Decimal by builder.defineDecimal("receive", receive, minimum = Decimal.ONE)
			override val extract: Decimal by builder.defineDecimal("extract", extract, minimum = Decimal.ONE)
		}

		builder.pop()

		return obj
	}

	private fun batteryValues(name: String, storage: Decimal, receive: Decimal, extract: Decimal = receive, initialBatteryLevel: Decimal = Decimal.ZERO): BatteryBalanceValues {
		builder.push(name)

		val obj = object : BatteryBalanceValues {
			override val capacity: Decimal by builder.defineDecimal("capacity", storage, minimum = Decimal.ONE)
			override val receive: Decimal by builder.defineDecimal("receive", receive, minimum = Decimal.ONE)
			override val extract: Decimal by builder.defineDecimal("extract", extract, minimum = Decimal.ONE)
			override val initialBatteryLevel: Decimal by builder.defineDecimal("initialBatteryLevel", initialBatteryLevel, minimum = Decimal.ZERO)
		}

		builder.pop()

		return obj
	}

	private fun conciseValues(name: String, storage: Decimal, throughput: Decimal): ConciseBalanceValues {
		builder.push(name)

		val obj = object : ConciseBalanceValues {
			override val capacity: Decimal by builder.defineDecimal("capacity", storage, minimum = Decimal.ONE)
			override val throughput: Decimal by builder.defineDecimal("throughput", throughput, minimum = Decimal.ONE)
		}

		builder.pop()

		return obj
	}

	init {
		builder.push("EnergyBatteries")
	}

	object Batteries {
		val CRUDE = batteryValues(MNames.BATTERY_CRUDE, Decimal(100_000), Decimal(160), Decimal(40), Decimal(80_000))
		val BASIC = batteryValues(MNames.BATTERY_BASIC, Decimal(400_000), Decimal(600))
		val NORMAL = batteryValues(MNames.BATTERY_NORMAL, Decimal(2_000_000), Decimal(1_000))
		val DENSE = batteryValues(MNames.BATTERY_DENSE, Decimal(10_000_000), Decimal(2_000))
		val CAPACITOR = batteryValues(MNames.BATTERY_CAPACITOR, Decimal(500_000), Decimal(50_000))

		val QUANTUM_BATTERY = conciseValues(MNames.QUANTUM_BATTERY, Decimal(40_000_000), Decimal(10_000))
		val QUANTUM_CAPACITOR = conciseValues(MNames.QUANTUM_CAPACITOR, Decimal(1_000_000), Decimal(200_000))

		val ZPM = conciseValues(MNames.ZPM_BATTERY, Decimal(200_000_000_000_000L), Decimal(200_000_000L))
	}

	init {
		Batteries
		builder.pop()
	}

	val MATTER_DUST_CAPACITY by builder.comment("Maximal matter value one matter dust item can have").defineDecimal("matterDustCapacity", Decimal(2_000), minimum = Decimal.ONE_TENTH)

	init {
		builder.push("MatterCapacitors")
	}

	object Capacitors {
		val BASIC by builder.defineDecimal(MNames.MATTER_CAPACITOR_BASIC, Decimal(2_500), minimum = Decimal.ONE_TENTH)
		val NORMAL by builder.defineDecimal(MNames.MATTER_CAPACITOR_NORMAL, Decimal(10_000), minimum = Decimal.ONE_TENTH)
		val DENSE by builder.defineDecimal(MNames.MATTER_CAPACITOR_DENSE, Decimal(25_000), minimum = Decimal.ONE_TENTH)
	}

	init {
		Capacitors
		builder.pop()
	}

	object PatternDrives {
		val NORMAL: Int by builder.defineInRange(MNames.PATTERN_DRIVE_NORMAL, 4, 1, Int.MAX_VALUE)
	}

	init {
		builder.push("PatternDrives")
		PatternDrives
		builder.pop()
	}

	val FLUID_CAPSULE_CAPACITY: Int by builder.defineInRange("FLUID_CAPSULE_CAPACITY", 1000, 1, Int.MAX_VALUE)
	val FLUID_TANK_CAPACITY: Int by builder.defineInRange("FLUID_TANK_CAPACITY", 32_000, 1, Int.MAX_VALUE)

	object ExplosiveHammer {
		val MAX_TRAVEL: Double by builder
			.comment("Max distance, in blocks, the nail can travel", "This also dictates potential max damage nail can deal (controlled by TRAVEL_DAMAGE_MULT)")
			.comment("The nail penetrate through blocks and entities, losing power in process")
			.comment("The nail can not travel into unloaded chunks and will stop immediately if it does so")
			.comment("The nail travels its entire path instantaneously")
			.defineInRange("MAX_TRAVEL", 48.0, 0.0, Double.MAX_VALUE)

		val TRAVEL_EXPLOSION_RESIST_MULT: Double by builder
			.comment("Multiplier of explosion resistance of blocks on nail path")
			.comment("If explosion resistance multiplied by this is bigger than nail current force (base of MAX_TRAVEL)")
			.comment("then nail movement halts, otherwise nail force is reduced by explosion resistance multiplied by this")
			.comment("and block is broken (if DAMAGE_BLOCKS is true, otherwise it only acts as stopping power)")
			.comment("---")
			.comment("Values over 1 will make blocks more resistant to be penetrated through (and broken)")
			.comment("Values below 1 will make blocks less resistant to be penetrated though (and broken)")
			.comment("Value of 0 means no matter how strong block resistance is, it will be broken by nail", "(this includes bedrock!!!)")
			.comment("---")
			.defineInRange("TRAVEL_EXPLOSION_RESIST_MULT", 1.75, 0.0, Double.MAX_VALUE)

		val TRAVEL_DAMAGE_MULT: Double by builder
			.comment("Multiplier of current nail force (base of MAX_TRAVEL) for damage values")
			.comment("That is, any being on path of nail will get damaged by nail power multiplied by this")
			.defineInRange("TRAVEL_DAMAGE_MULT", 0.6, 0.0, Double.MAX_VALUE)

		val TRAVEL_MAX_HEALTH_MULT: Double by builder
			.comment("Multiplier of being's max health in nail path subtracted from nail force when penetrating through")
			.defineInRange("TRAVEL_MAX_HEALTH_MULT", 0.4, 0.0, Double.MAX_VALUE)

		val DAMAGE_BLOCKS: Boolean by builder
			.comment("Should hammer (**not** explosion) damage blocks", "Blocks damaged this way will always be dropped")
			.define("DAMAGE_BLOCKS", true)

		val EXPLOSION_DAMAGE_BLOCKS: Boolean by builder
			.comment("Should hammer **explosion** damage blocks")
			.define("EXPLOSION_DAMAGE_BLOCKS", false)

		val FLY_OFF_CHANCE: Double by builder
			.comment("Chance that hammer will fly off hands of attacker upon impact")
			.defineInRange("FLY_OFF_CHANCE", 0.6, 0.0, 1.0)

		val SELF_HARM_CHANCE: Double by builder
			.comment("When hammer DOES NOT fly off hands, what is chance of it damaging the user?", "(this does not account for damage because of hitting things close)", "(idiot)")
			.defineInRange("SELF_HARM_CHANCE", 0.4, 0.0, 1.0)

		val SELF_HARM_MIN_DAMAGE: Double by builder
			.comment("When hammer DOES NOT fly off hands, what is minimal damage it deals to the user (in half of hearts)?")
			.defineInRange("SELF_HARM_MIN_DAMAGE", 2.0, 0.0, Double.MAX_VALUE)

		val SELF_HARM_MAX_DAMAGE: Double by builder
			.comment("When hammer DOES NOT fly off hands, what is maximal damage it deals to the user (in half of hearts)?")
			.defineInRange("SELF_HARM_MAX_DAMAGE", 10.0, 0.0, Double.MAX_VALUE)

		val FLY_OFF_DAMAGE_CHANCE: Double by builder
			.comment("When hammer flies off hands, what is chance of it damaging the user?")
			.defineInRange("FLY_OFF_DAMAGE_CHANCE", 0.8, 0.0, 1.0)

		val FLY_OFF_MIN_DAMAGE: Double by builder
			.comment("When hammer flies off hands, what is minimal damage it deals to the user (in half of hearts)?")
			.defineInRange("FLY_OFF_MIN_DAMAGE", 4.0, 0.0, Double.MAX_VALUE)

		val FLY_OFF_MAX_DAMAGE: Double by builder
			.comment("When hammer flies off hands, what is maximal damage it deals to the user (in half of hearts)?")
			.defineInRange("FLY_OFF_MAX_DAMAGE", 25.0, 0.0, Double.MAX_VALUE)
	}

	init {
		builder.push("ExplosiveHammer")
		ExplosiveHammer
		builder.pop()
	}
}
